import unittest
from unittest import mock
from leeger.enum.MatchupType import MatchupType
from leeger.model.league.Division import Division

from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year

from leeger.league_loader.FleaflickerLeagueLoader import FleaflickerLeagueLoader


class TestFleaflickerLeagueLoader(unittest.TestCase):
    def test_init_leagueIdNotIntConvertable_raisesException(self):
        with self.assertRaises(ValueError) as context:
            FleaflickerLeagueLoader("foo", [])
        self.assertEqual("League ID 'foo' could not be turned into an int.", str(context.exception))

    @mock.patch("fleaflicker.api.LeagueInfoAPIClient.LeagueInfoAPIClient.get_league_standings")
    @mock.patch("fleaflicker.api.ScoringAPIClient.ScoringAPIClient.get_league_scoreboard")
    def test_loadLeague_happyPath(self, mockGetLeagueScoreboard, mockGetLeaguestandings):
        mockTeam1_2022 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2022 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2022 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2022 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2022 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2022 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2022 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2022 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockTeam1_2023 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2023 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2023 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2023 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2023 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2023 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2023 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2023 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockLeagueStandings2022 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2022",
                    "teams": [mockTeam1_2022, mockTeam2_2022, mockTeam3_2022, mockTeam4_2022],
                },
                {
                    "id": 2,
                    "name": "d2_2022",
                    "teams": [mockTeam5_2022, mockTeam6_2022, mockTeam7_2022, mockTeam8_2022],
                },
            ],
            "league": {"name": "Test League 2022", "id": 123},
            "season": 2022,
        }

        mockLeagueStandings2023 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2023",
                    "teams": [
                        mockTeam1_2023,
                        mockTeam2_2023,
                        mockTeam3_2023,
                        mockTeam4_2023,
                        mockTeam5_2023,
                        mockTeam6_2023,
                        mockTeam7_2023,
                        mockTeam8_2023,
                    ],
                }
            ],
            "league": {"name": "Test League 2023", "id": 123},
            "season": 2023,
        }

        mockWeek1_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2022,
                    "home": mockTeam4_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2022,
                    "home": mockTeam6_2022,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2022,
                    "home": mockTeam8_2022,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek1_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2023,
                    "home": mockTeam4_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2023,
                    "home": mockTeam6_2023,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2023,
                    "home": mockTeam8_2023,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek2_2022 = {
            "games": [
                {
                    "away": mockTeam2_2022,
                    "home": mockTeam3_2022,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek2_2023 = {
            "games": [
                {
                    "away": mockTeam2_2023,
                    "home": mockTeam3_2023,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek3_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockWeek3_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockScoreboard2022 = {
            "eligibleSchedulePeriods": [mockWeek1_2022, mockWeek2_2022, mockWeek3_2022]
        }

        mockScoreboard2023 = {
            "eligibleSchedulePeriods": [mockWeek1_2023, mockWeek2_2023, mockWeek3_2023]
        }

        mockGetLeaguestandings.side_effect = [mockLeagueStandings2022, mockLeagueStandings2023]
        mockGetLeagueScoreboard.side_effect = [
            mockScoreboard2022,
            mockWeek1_2022,
            mockWeek2_2022,
            mockWeek3_2022,
            mockScoreboard2023,
            mockWeek1_2023,
            mockWeek2_2023,
            mockWeek3_2023,
        ]

        leagueLoader = FleaflickerLeagueLoader("123", [2022, 2023])
        league = leagueLoader.loadLeague()

        # expected league
        division1_2022 = Division(name="d1_2022")
        division2_2022 = Division(name="d2_2022")

        division1_2023 = Division(name="d1_2023")

        owner1 = Owner(name="Owner 1")
        owner2 = Owner(name="Owner 2")
        owner3 = Owner(name="Owner 3")
        owner4 = Owner(name="Owner 4")
        owner5 = Owner(name="Owner 5")
        owner6 = Owner(name="Owner 6")
        owner7 = Owner(name="Owner 7")
        owner8 = Owner(name="Owner 8")

        team1_2022 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2022.id)
        team2_2022 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2022.id)
        team3_2022 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2022.id)
        team4_2022 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2022.id)
        team5_2022 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2022.id)
        team6_2022 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2022.id)
        team7_2022 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2022.id)
        team8_2022 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2022.id)

        team1_2023 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2023.id)
        team2_2023 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2023.id)
        team3_2023 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2023.id)
        team4_2023 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2023.id)
        team5_2023 = Team(ownerId=owner5.id, name="Team 5", divisionId=division1_2023.id)
        team6_2023 = Team(ownerId=owner6.id, name="Team 6", divisionId=division1_2023.id)
        team7_2023 = Team(ownerId=owner7.id, name="Team 7", divisionId=division1_2023.id)
        team8_2023 = Team(ownerId=owner8.id, name="Team 8", divisionId=division1_2023.id)

        expectedLeague = League(
            name="Test League 2023",
            owners=[owner1, owner2, owner3, owner4, owner5, owner6, owner7, owner8],
            years=[
                Year(
                    yearNumber=2022,
                    teams=[
                        team1_2022,
                        team2_2022,
                        team3_2022,
                        team4_2022,
                        team5_2022,
                        team6_2022,
                        team7_2022,
                        team8_2022,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2022.id,
                                    teamBId=team4_2022.id,
                                    teamAScore=100,
                                    teamBScore=90,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2022.id,
                                    teamBId=team6_2022.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2022.id,
                                    teamBId=team8_2022.id,
                                    teamAScore=100.2,
                                    teamBScore=90.2,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2022.id,
                                    teamBId=team3_2022.id,
                                    teamAScore=100.4,
                                    teamBScore=90.4,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100.5,
                                    teamBScore=90.5,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2022, division2_2022],
                ),
                Year(
                    yearNumber=2023,
                    teams=[
                        team1_2023,
                        team2_2023,
                        team3_2023,
                        team4_2023,
                        team5_2023,
                        team6_2023,
                        team7_2023,
                        team8_2023,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=100,
                                    teamBScore=90,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2023.id,
                                    teamBId=team6_2023.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2023.id,
                                    teamBId=team8_2023.id,
                                    teamAScore=100.2,
                                    teamBScore=90.2,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team3_2023.id,
                                    teamAScore=100.4,
                                    teamBScore=90.4,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100.5,
                                    teamBScore=90.5,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2023],
                ),
            ],
        )

        self.assertTrue(league.equals(expectedLeague, ignoreBaseIds=True, ignoreIds=True))
        # check multiWeekMatchupIds
        for year in league.years:
            for week in year.weeks:
                for matchup in week.matchups:
                    self.assertIsNone(matchup.multiWeekMatchupId)

    @mock.patch("fleaflicker.api.LeagueInfoAPIClient.LeagueInfoAPIClient.get_league_standings")
    @mock.patch("fleaflicker.api.ScoringAPIClient.ScoringAPIClient.get_league_scoreboard")
    def test_loadLeague_happyPath_withOwnerNamesAndAliases(
        self, mockGetLeagueScoreboard, mockGetLeaguestandings
    ):
        mockTeam1_2022 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2022 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2022 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2022 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2022 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2022 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2022 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2022 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockTeam1_2023 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2023 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2023 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2023 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2023 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2023 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2023 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2023 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockLeagueStandings2022 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2022",
                    "teams": [mockTeam1_2022, mockTeam2_2022, mockTeam3_2022, mockTeam4_2022],
                },
                {
                    "id": 2,
                    "name": "d2_2022",
                    "teams": [mockTeam5_2022, mockTeam6_2022, mockTeam7_2022, mockTeam8_2022],
                },
            ],
            "league": {"name": "Test League 2022", "id": 123},
            "season": 2022,
        }

        mockLeagueStandings2023 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2023",
                    "teams": [
                        mockTeam1_2023,
                        mockTeam2_2023,
                        mockTeam3_2023,
                        mockTeam4_2023,
                        mockTeam5_2023,
                        mockTeam6_2023,
                        mockTeam7_2023,
                        mockTeam8_2023,
                    ],
                }
            ],
            "league": {"name": "Test League 2023", "id": 123},
            "season": 2023,
        }

        mockWeek1_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2022,
                    "home": mockTeam4_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2022,
                    "home": mockTeam6_2022,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2022,
                    "home": mockTeam8_2022,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek1_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2023,
                    "home": mockTeam4_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2023,
                    "home": mockTeam6_2023,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2023,
                    "home": mockTeam8_2023,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek2_2022 = {
            "games": [
                {
                    "away": mockTeam2_2022,
                    "home": mockTeam3_2022,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek2_2023 = {
            "games": [
                {
                    "away": mockTeam2_2023,
                    "home": mockTeam3_2023,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek3_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockWeek3_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockScoreboard2022 = {
            "eligibleSchedulePeriods": [mockWeek1_2022, mockWeek2_2022, mockWeek3_2022]
        }

        mockScoreboard2023 = {
            "eligibleSchedulePeriods": [mockWeek1_2023, mockWeek2_2023, mockWeek3_2023]
        }

        mockGetLeaguestandings.side_effect = [mockLeagueStandings2022, mockLeagueStandings2023]
        mockGetLeagueScoreboard.side_effect = [
            mockScoreboard2022,
            mockWeek1_2022,
            mockWeek2_2022,
            mockWeek3_2022,
            mockScoreboard2023,
            mockWeek1_2023,
            mockWeek2_2023,
            mockWeek3_2023,
        ]

        leagueLoader = FleaflickerLeagueLoader(
            "123",
            [2022, 2023],
            ownerNamesAndAliases={
                "o1": ["Owner 1"],
                "o2": ["Owner 2"],
                "o3": ["Owner 3"],
                "o4": ["Owner 4"],
                "o5": ["Owner 5"],
                "o6": ["Owner 6"],
                "o7": ["Owner 7"],
                "o8": ["Owner 8"],
            },
        )
        league = leagueLoader.loadLeague()

        # expected league
        division1_2022 = Division(name="d1_2022")
        division2_2022 = Division(name="d2_2022")

        division1_2023 = Division(name="d1_2023")

        owner1 = Owner(name="o1")
        owner2 = Owner(name="o2")
        owner3 = Owner(name="o3")
        owner4 = Owner(name="o4")
        owner5 = Owner(name="o5")
        owner6 = Owner(name="o6")
        owner7 = Owner(name="o7")
        owner8 = Owner(name="o8")

        team1_2022 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2022.id)
        team2_2022 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2022.id)
        team3_2022 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2022.id)
        team4_2022 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2022.id)
        team5_2022 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2022.id)
        team6_2022 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2022.id)
        team7_2022 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2022.id)
        team8_2022 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2022.id)

        team1_2023 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2023.id)
        team2_2023 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2023.id)
        team3_2023 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2023.id)
        team4_2023 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2023.id)
        team5_2023 = Team(ownerId=owner5.id, name="Team 5", divisionId=division1_2023.id)
        team6_2023 = Team(ownerId=owner6.id, name="Team 6", divisionId=division1_2023.id)
        team7_2023 = Team(ownerId=owner7.id, name="Team 7", divisionId=division1_2023.id)
        team8_2023 = Team(ownerId=owner8.id, name="Team 8", divisionId=division1_2023.id)

        expectedLeague = League(
            name="Test League 2023",
            owners=[owner1, owner2, owner3, owner4, owner5, owner6, owner7, owner8],
            years=[
                Year(
                    yearNumber=2022,
                    teams=[
                        team1_2022,
                        team2_2022,
                        team3_2022,
                        team4_2022,
                        team5_2022,
                        team6_2022,
                        team7_2022,
                        team8_2022,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2022.id,
                                    teamBId=team4_2022.id,
                                    teamAScore=100,
                                    teamBScore=90,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2022.id,
                                    teamBId=team6_2022.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2022.id,
                                    teamBId=team8_2022.id,
                                    teamAScore=100.2,
                                    teamBScore=90.2,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2022.id,
                                    teamBId=team3_2022.id,
                                    teamAScore=100.4,
                                    teamBScore=90.4,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100.5,
                                    teamBScore=90.5,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2022, division2_2022],
                ),
                Year(
                    yearNumber=2023,
                    teams=[
                        team1_2023,
                        team2_2023,
                        team3_2023,
                        team4_2023,
                        team5_2023,
                        team6_2023,
                        team7_2023,
                        team8_2023,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=100,
                                    teamBScore=90,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2023.id,
                                    teamBId=team6_2023.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2023.id,
                                    teamBId=team8_2023.id,
                                    teamAScore=100.2,
                                    teamBScore=90.2,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team3_2023.id,
                                    teamAScore=100.4,
                                    teamBScore=90.4,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100.5,
                                    teamBScore=90.5,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2023],
                ),
            ],
        )

        self.assertTrue(league.equals(expectedLeague, ignoreBaseIds=True, ignoreIds=True))
        # check multiWeekMatchupIds
        for year in league.years:
            for week in year.weeks:
                for matchup in week.matchups:
                    self.assertIsNone(matchup.multiWeekMatchupId)

    @mock.patch("fleaflicker.api.LeagueInfoAPIClient.LeagueInfoAPIClient.get_league_standings")
    @mock.patch("fleaflicker.api.ScoringAPIClient.ScoringAPIClient.get_league_scoreboard")
    def test_loadLeague_withLeagueName(self, mockGetLeagueScoreboard, mockGetLeaguestandings):
        mockTeam1_2022 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2022 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2022 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2022 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2022 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2022 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2022 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2022 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockTeam1_2023 = {"owners": [{"displayName": "Owner 1"}], "id": 1, "name": "Team 1"}
        mockTeam2_2023 = {"owners": [{"displayName": "Owner 2"}], "id": 2, "name": "Team 2"}
        mockTeam3_2023 = {"owners": [{"displayName": "Owner 3"}], "id": 3, "name": "Team 3"}
        mockTeam4_2023 = {"owners": [{"displayName": "Owner 4"}], "id": 4, "name": "Team 4"}
        mockTeam5_2023 = {"owners": [{"displayName": "Owner 5"}], "id": 5, "name": "Team 5"}
        mockTeam6_2023 = {"owners": [{"displayName": "Owner 6"}], "id": 6, "name": "Team 6"}
        mockTeam7_2023 = {"owners": [{"displayName": "Owner 7"}], "id": 7, "name": "Team 7"}
        mockTeam8_2023 = {"owners": [{"displayName": "Owner 8"}], "id": 8, "name": "Team 8"}

        mockLeagueStandings2022 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2022",
                    "teams": [mockTeam1_2022, mockTeam2_2022, mockTeam3_2022, mockTeam4_2022],
                },
                {
                    "id": 2,
                    "name": "d2_2022",
                    "teams": [mockTeam5_2022, mockTeam6_2022, mockTeam7_2022, mockTeam8_2022],
                },
            ],
            "league": {"name": "Test League 2022", "id": 123},
            "season": 2022,
        }

        mockLeagueStandings2023 = {
            "divisions": [
                {
                    "id": 1,
                    "name": "d1_2023",
                    "teams": [
                        mockTeam1_2023,
                        mockTeam2_2023,
                        mockTeam3_2023,
                        mockTeam4_2023,
                        mockTeam5_2023,
                        mockTeam6_2023,
                        mockTeam7_2023,
                        mockTeam8_2023,
                    ],
                }
            ],
            "league": {"name": "Test League 2023", "id": 123},
            "season": 2023,
        }

        mockWeek1_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2022,
                    "home": mockTeam4_2022,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2022,
                    "home": mockTeam6_2022,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2022,
                    "home": mockTeam8_2022,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek1_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 100}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam3_2023,
                    "home": mockTeam4_2023,
                    "awayScore": {"score": {"value": 100}},
                    "homeScore": {"score": {"value": 90}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam5_2023,
                    "home": mockTeam6_2023,
                    "awayScore": {"score": {"value": 100.1}},
                    "homeScore": {"score": {"value": 90.1}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
                {
                    "away": mockTeam7_2023,
                    "home": mockTeam8_2023,
                    "awayScore": {"score": {"value": 100.2}},
                    "homeScore": {"score": {"value": 90.2}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                },
            ]
        }

        mockWeek2_2022 = {
            "games": [
                {
                    "away": mockTeam2_2022,
                    "home": mockTeam3_2022,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek2_2023 = {
            "games": [
                {
                    "away": mockTeam2_2023,
                    "home": mockTeam3_2023,
                    "awayScore": {"score": {"value": 100.4}},
                    "homeScore": {"score": {"value": 90.4}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isThirdPlaceGame": True,
                }
            ]
        }

        mockWeek3_2022 = {
            "games": [
                {
                    "away": mockTeam1_2022,
                    "home": mockTeam2_2022,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockWeek3_2023 = {
            "games": [
                {
                    "away": mockTeam1_2023,
                    "home": mockTeam2_2023,
                    "awayScore": {"score": {"value": 100.5}},
                    "homeScore": {"score": {"value": 90.5}},
                    "awayResult": "WIN",
                    "homeResult": "LOSS",
                    "isFinalScore": True,
                    "isChampionshipGame": True,
                }
            ]
        }

        mockScoreboard2022 = {
            "eligibleSchedulePeriods": [mockWeek1_2022, mockWeek2_2022, mockWeek3_2022]
        }

        mockScoreboard2023 = {
            "eligibleSchedulePeriods": [mockWeek1_2023, mockWeek2_2023, mockWeek3_2023]
        }

        mockGetLeaguestandings.side_effect = [mockLeagueStandings2022, mockLeagueStandings2023]
        mockGetLeagueScoreboard.side_effect = [
            mockScoreboard2022,
            mockWeek1_2022,
            mockWeek2_2022,
            mockWeek3_2022,
            mockScoreboard2023,
            mockWeek1_2023,
            mockWeek2_2023,
            mockWeek3_2023,
        ]

        leagueLoader = FleaflickerLeagueLoader("123", [2022, 2023], leagueName="custom name")
        league = leagueLoader.loadLeague()

        self.assertEqual("custom name", league.name)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from leeger.model.abstract.EqualityCheck import EqualityCheck

from leeger.model.abstract.UniqueId import UniqueId
from leeger.util.CustomLogger import CustomLogger
from leeger.util.JSONDeserializable import JSONDeserializable
from leeger.util.JSONSerializable import JSONSerializable
from leeger.util.equality import modelEquals


@dataclass(kw_only=True, eq=False)
class Team(UniqueId, EqualityCheck, JSONSerializable, JSONDeserializable):
    __LOGGER = CustomLogger.getLogger()
    ownerId: str
    name: str
    divisionId: Optional[str] = None

    def equals(
        self,
        otherTeam: Team,
        *,
        ignoreIds: bool = False,
        ignoreBaseIds: bool = False,
        logDifferences: bool = False,
    ) -> bool:
        """
        Checks if *this* Team is the same as the given Team.
        """

        return modelEquals(
            objA=self,
            objB=otherTeam,
            baseFields={"name"},
            idFields={"ownerId", "divisionId"},
            parentKey="Team",
            ignoreIdFields=ignoreIds,
            ignoreBaseIdField=ignoreBaseIds,
            logDifferences=logDifferences,
        )

    def __eq__(self, otherTeam: Team) -> bool:
        self.__LOGGER.info("Use .equals() for more options when comparing Team instances.")
        return self.equals(otherTeam=otherTeam)

    def toJson(self) -> dict:
        return {
            "id": self.id,
            "ownerId": self.ownerId,
            "name": self.name,
            "divisionId": self.divisionId,
        }

    @staticmethod
    def fromJson(d: dict) -> Team:
        team = Team(ownerId=d["ownerId"], name=d["name"], divisionId=d.get("divisionId"))
        team.id = d["id"]
        return team



from abc import abstractmethod
from typing import Optional

from leeger.exception.DoesNotExistException import DoesNotExistException
from leeger.exception.LeagueLoaderException import LeagueLoaderException
from leeger.model.league.League import League
from leeger.model.league.Owner import Owner
from leeger.model.league.Year import Year
from leeger.util.CustomLogger import CustomLogger


class LeagueLoader:
    """
    League Loader classes should inherit this.
    The point of a league loader is to load a League object from different Fantasy Football sources.
    """

    def __init__(
        self,
        leagueId: str,
        years: list[int],
        *,
        ownerNamesAndAliases: Optional[dict] = None,
        leagueName: Optional[str] = None,
    ):
        self._LOGGER = CustomLogger().getLogger()
        # validation
        if len(years) == 0:
            raise ValueError(f"No years given to load league with ID '{leagueId}'.")

        if not all(isinstance(year, int) for year in years):
            raise ValueError(f"All given years must be ints.")

        self._leagueId = leagueId
        self._years = sorted(years)
        self._owners: Optional[list[Owner]] = None
        # owners may have multiple names across different years,
        # defining owner names and aliases allows users to have multiple names that can belong to the same owner.
        # this prevents issues where an owner with a name change across years is counted as 2 different owners.
        # this should be formatted like so:
        # ownerNamesAndAliases = {"someOwnerNameIWant": ["alias1", "alias2"],
        #                           someOtherOwnerNameIWant: ["alias3", "alias4"]}
        self._ownerNamesAndAliases: dict[str, list[str]] = (
            ownerNamesAndAliases if ownerNamesAndAliases else dict()
        )
        self._leagueName = leagueName
        self._leagueNameByYear: dict[
            int, str
        ] = dict()  # will hold league name by year like {2020: "foo", 2021: "baz", ...}

    def _getLeagueName(self) -> str:
        leagueName = self._leagueName
        if self._leagueName is None:
            if len(self._leagueNameByYear.keys()) == 0:
                raise LeagueLoaderException(
                    "Tried to retrieve league name with no leagueName parameter given and no league names set."
                )
            # use most recent year league name
            mostRecentYear = sorted(self._leagueNameByYear.keys())[-1]
            leagueName = self._leagueNameByYear[mostRecentYear]
        return leagueName

    def _getValidYears(self, years: list[Year]) -> list[Year]:
        validYears = list()
        # make sure years are ordered oldest -> newest
        for year in sorted(years, key=lambda y: y.yearNumber):
            if len(year.weeks) > 0:
                validYears.append(year)
            else:
                self._LOGGER.warning(
                    f"Year '{year.yearNumber}' discarded for not having any weeks."
                )
        return validYears

    def _validateRetrievedLeagues(self, retrievedLeagues: list) -> None:
        expectedLeagueCount = len(self._years)
        actualLeagueCount = len(retrievedLeagues)
        if actualLeagueCount != expectedLeagueCount:
            raise LeagueLoaderException(
                f"Expected to retrieve {expectedLeagueCount} league/s, got {actualLeagueCount} league/s."
            )

    def _getGeneralOwnerNameFromGivenOwnerName(self, givenOwnerName: str) -> Optional[str]:
        foundGeneralOwnerName = None
        for generalOwnerName, aliases in self._ownerNamesAndAliases.items():
            if givenOwnerName in aliases:
                foundGeneralOwnerName = generalOwnerName
                break
        return foundGeneralOwnerName

    def _getOwnerByName(self, ownerName: str) -> Owner:
        generalOwnerName = self._getGeneralOwnerNameFromGivenOwnerName(ownerName)
        for owner in self._owners:
            if ownerName == owner.name or generalOwnerName == owner.name:
                return owner
        raise DoesNotExistException(
            f"Owner name '{ownerName}' does not match any previously loaded owner names. To add multiple names for a single owner, use the 'ownerNamesAndAliases' keyword argument to define them."
        )

    ## validation

    def _warnForUnusedOwnerNames(self, league: League) -> None:
        """
        Logs a warning for all owner names that are not in the loaded league IF ownerNamesAndAliases is given.
        """
        if self._ownerNamesAndAliases:
            allLoadedOwnerNames = [owner.name for owner in league.owners]
            unusedOwnerNames = list()
            for ownerName in self._ownerNamesAndAliases.keys():
                if ownerName not in allLoadedOwnerNames:
                    unusedOwnerNames.append(ownerName)
            if unusedOwnerNames:
                self._LOGGER.warning(
                    f"Some owner names were given but not assigned to the loaded League: {unusedOwnerNames}"
                )

    @abstractmethod
    def loadLeague(self, validate: bool = True, *args, **kwargs) -> League:
        ...

    @abstractmethod
    def getOwnerNames(self, *args, **kwargs) -> dict[int, list[str]]:
        ...



import itertools
from typing import Optional

from sleeper.api import LeagueAPIClient
from sleeper.enum import Sport as SleeperSport
from sleeper.enum import SeasonStatus as SleeperSeasonStatus
from sleeper.enum import PlayoffRoundType as SleeperPlayoffRoundType
from sleeper.model import League as SleeperLeague
from sleeper.model import Matchup as SleeperMatchup
from sleeper.model import PlayoffMatchup as SleeperPlayoffMatchup
from sleeper.model import User as SleeperUser
from sleeper.model import SportState as SleeperSportState

from leeger.enum.MatchupType import MatchupType
from leeger.exception.DoesNotExistException import DoesNotExistException
from leeger.league_loader.LeagueLoader import LeagueLoader
from leeger.model.league import YearSettings
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.validate import leagueValidation
from leeger.exception.LeagueLoaderException import LeagueLoaderException
from leeger.model.league.Division import Division


class SleeperLeagueLoader(LeagueLoader):
    """
    Responsible for loading a League from Sleeper Fantasy Football.
    https://sleeper.com/
    """

    __INVALID_SLEEPER_LEAGUE_IDS = [None, "0"]

    def __init__(
        self,
        mostRecentLeagueId: str,
        years: list[int],
        *,
        ownerNamesAndAliases: Optional[dict[str, list[str]]] = None,
        leagueName: Optional[str] = None,
    ):
        super().__init__(
            mostRecentLeagueId,
            years,
            ownerNamesAndAliases=ownerNamesAndAliases,
            leagueName=leagueName,
        )

        self.__sleeperUserIdToOwnerMap: dict[str, Owner] = dict()
        self.__sleeperRosterIdToTeamMap: dict[int, Team] = dict()
        self.__SLEEPER_USERS_BY_LEAGUE_ID_CACHE = dict()  # functions as a cache for Sleeper Users
        self.__SLEEPER_SPORT_STATE_CACHE: SleeperSportState = (
            None  # functions as a cache for Sleeper SportState
        )
        self.__sleeperDivisionIdToDivisionMap: dict[
            int, Division
        ] = dict()  # holds the division info for ONLY the current year

    def __resetCaches(self) -> None:
        self.__SLEEPER_USERS_BY_LEAGUE_ID_CACHE = dict()
        self.__SLEEPER_SPORT_STATE_CACHE = None

    def __getSleeperUsers(self, leagueId: str) -> list[SleeperUser]:
        if leagueId not in self.__SLEEPER_USERS_BY_LEAGUE_ID_CACHE:
            # don't have these users loaded yet
            sleeperUsers = LeagueAPIClient.get_users_in_league(league_id=leagueId)
            self.__SLEEPER_USERS_BY_LEAGUE_ID_CACHE[leagueId] = sleeperUsers
            return sleeperUsers
        # do have these users loaded
        return self.__SLEEPER_USERS_BY_LEAGUE_ID_CACHE[leagueId]

    def __getSleeperSportState(self):
        if self.__SLEEPER_SPORT_STATE_CACHE is None:
            self.__SLEEPER_SPORT_STATE_CACHE = LeagueAPIClient.get_sport_state(
                sport=SleeperSport.NFL
            )
        return self.__SLEEPER_SPORT_STATE_CACHE

    def __getAllLeagues(self) -> list[SleeperLeague]:
        sleeperLeagues = list()
        years = self._years.copy()
        currentLeagueId = self._leagueId
        while len(years) > 0 and currentLeagueId not in self.__INVALID_SLEEPER_LEAGUE_IDS:
            currentLeague: SleeperLeague = LeagueAPIClient.get_league(league_id=currentLeagueId)
            if int(currentLeague.season) in years:
                # we only want to add valid seasons
                # NOTE: Not sure if we should include SleeperSeasonStatus.POSTPONED here or not
                if currentLeague.status not in (
                    SleeperSeasonStatus.COMPLETE,
                    SleeperSeasonStatus.IN_SEASON,
                    SleeperSeasonStatus.POST_SEASON,
                ):
                    raise LeagueLoaderException(
                        f"Year {currentLeague.season} has a status that is not supported: '{currentLeague.status}'"
                    )
                sleeperLeagues.append(currentLeague)
                years.remove(int(currentLeague.season))
            currentLeagueId = currentLeague.previous_league_id

        # reverse list so most recent is last in list
        sleeperLeagues = sleeperLeagues[::-1]
        self._validateRetrievedLeagues(sleeperLeagues)
        return sleeperLeagues

    def getOwnerNames(self) -> dict[int, list[str]]:
        yearToOwnerNamesMap: dict[int, list[str]] = dict()
        sleeperLeagues = self.__getAllLeagues()
        for sleeperLeague in sleeperLeagues:
            yearToOwnerNamesMap[int(sleeperLeague.season)] = list()
            sleeperUsers = self.__getSleeperUsers(sleeperLeague.league_id)
            for sleeperUser in sleeperUsers:
                ownerName = sleeperUser.display_name
                yearToOwnerNamesMap[int(sleeperLeague.season)].append(ownerName)
        return yearToOwnerNamesMap

    def loadLeague(self, validate: bool = True) -> League:
        sleeperLeagues = self.__getAllLeagues()
        league = self.__buildLeague(sleeperLeagues)
        if validate:
            # validate new league
            leagueValidation.runAllChecks(league)
        self.__resetCaches()
        self._warnForUnusedOwnerNames(league)
        return league

    def __buildLeague(self, sleeperLeagues: list[SleeperLeague]) -> League:
        years = list()
        self.__loadOwners(sleeperLeagues)
        owners = list(self.__sleeperUserIdToOwnerMap.values())
        for sleeperLeague in sleeperLeagues:
            # save league name for each year
            self._leagueNameByYear[int(sleeperLeague.season)] = sleeperLeague.name
            years.append(self.__buildYear(sleeperLeague))
        return League(name=self._getLeagueName(), owners=owners, years=self._getValidYears(years))

    def __buildYear(self, sleeperLeague: SleeperLeague) -> Year:
        # save division info if applicable
        if self.__yearHasDivisions(sleeperLeague):
            for divisionNumber in range(1, sleeperLeague.settings.divisions + 1):
                self.__sleeperDivisionIdToDivisionMap[divisionNumber] = Division(
                    name=getattr(sleeperLeague.metadata, f"division_{divisionNumber}")
                )
        teams = self.__buildTeams(sleeperLeague)
        weeks = self.__buildWeeks(sleeperLeague)
        # add YearSettings
        yearSettings = YearSettings()
        if sleeperLeague.settings.league_average_match == 1:
            yearSettings.leagueMedianGames = True

        year = Year(
            yearNumber=int(sleeperLeague.season),
            teams=teams,
            weeks=weeks,
            yearSettings=yearSettings,
            divisions=None
            if not self.__yearHasDivisions(sleeperLeague)
            else list(self.__sleeperDivisionIdToDivisionMap.values()),
        )
        # clear division info
        self.__sleeperDivisionIdToDivisionMap = dict()
        return year

    def __buildWeeks(self, sleeperLeague: SleeperLeague) -> list[Week]:
        weeks = list()
        # get regular season weeks
        # once we have found an incomplete week, all weeks after will also be incomplete
        foundIncompleteWeek = False
        for weekNumber in range(1, sleeperLeague.settings.playoff_week_start):
            if not foundIncompleteWeek and self.__isCompletedWeek(weekNumber, sleeperLeague):
                # get each teams matchup for that week
                matchups = list()
                sleeperMatchupsForThisWeek = LeagueAPIClient.get_matchups_for_week(
                    league_id=sleeperLeague.league_id, week=weekNumber
                )
                sleeperMatchupIdToSleeperMatchupMap: dict[int, list[SleeperMatchup]] = dict()
                for sleeperMatchup in sleeperMatchupsForThisWeek:
                    if sleeperMatchup.matchup_id in sleeperMatchupIdToSleeperMatchupMap.keys():
                        sleeperMatchupIdToSleeperMatchupMap[sleeperMatchup.matchup_id].append(
                            sleeperMatchup
                        )
                    else:
                        sleeperMatchupIdToSleeperMatchupMap[sleeperMatchup.matchup_id] = [
                            sleeperMatchup
                        ]

                for sleeperMatchupPair in sleeperMatchupIdToSleeperMatchupMap.values():
                    # team A
                    teamASleeperMatchup = sleeperMatchupPair[0]
                    teamA = self.__sleeperRosterIdToTeamMap[teamASleeperMatchup.roster_id]

                    # team B
                    teamBSleeperMatchup = sleeperMatchupPair[1]
                    teamB = self.__sleeperRosterIdToTeamMap[teamBSleeperMatchup.roster_id]

                    # sleeper does not have tiebreakers for regular season games
                    # Source: https://support.sleeper.app/en/articles/4238872-can-i-set-tiebreakers#:~:text=We%20do%20not%20offer%20any,and%20adjust%20the%20point%20total.
                    matchups.append(
                        Matchup(
                            teamAId=teamA.id,
                            teamBId=teamB.id,
                            teamAScore=teamASleeperMatchup.points,
                            teamBScore=teamBSleeperMatchup.points,
                            teamAHasTiebreaker=False,
                            teamBHasTiebreaker=False,
                            matchupType=MatchupType.REGULAR_SEASON,
                        )
                    )
                weeks.append(Week(weekNumber=weekNumber, matchups=matchups))
            else:
                foundIncompleteWeek = True
        # get playoff weeks
        # NOTE: bye weeks will not be returned here. That's ok because we don't want those anyways
        allSleeperPlayoffMatchups = LeagueAPIClient.get_winners_bracket(
            league_id=sleeperLeague.league_id
        )
        if len(allSleeperPlayoffMatchups) > 0:
            # sort sleeperPlayoffMatchups by round into a dict
            playoffRoundAndSleeperPlayoffMatchups: dict[int, list[SleeperPlayoffMatchup]] = dict()
            for sleeperPlayoffMatchup in allSleeperPlayoffMatchups:
                if sleeperPlayoffMatchup.round in playoffRoundAndSleeperPlayoffMatchups.keys():
                    playoffRoundAndSleeperPlayoffMatchups[sleeperPlayoffMatchup.round].append(
                        sleeperPlayoffMatchup
                    )
                else:
                    playoffRoundAndSleeperPlayoffMatchups[sleeperPlayoffMatchup.round] = [
                        sleeperPlayoffMatchup
                    ]
            numberOfPlayoffRounds = max(
                [playoffMatchup.round for playoffMatchup in allSleeperPlayoffMatchups]
            )  # don't know a better way to determine this
            numberOfPlayoffWeeks = self.__calculate_number_of_playoff_weeks(
                sleeperLeague, allSleeperPlayoffMatchups
            )
            playoffWeeks = list(
                range(
                    sleeperLeague.settings.playoff_week_start,
                    sleeperLeague.settings.playoff_week_start + numberOfPlayoffWeeks,
                )
            )
            playoffWeekRoundList = self.__create_playoff_week_round_list(
                sleeperLeague, playoffWeeks, numberOfPlayoffRounds
            )
            for weekNumber, roundNumber in playoffWeekRoundList:
                # get each teams matchup for that week
                matchups = list()
                sleeperMatchupsForThisWeek = LeagueAPIClient.get_matchups_for_week(
                    league_id=sleeperLeague.league_id, week=weekNumber
                )
                # remove matchups that don't have a matchup id
                sleeperMatchupsForThisWeek = [
                    sleeperMatchup
                    for sleeperMatchup in sleeperMatchupsForThisWeek
                    if sleeperMatchup.matchup_id is not None
                ]
                # used to check if a Sleeper playoff matchup is in this week's matchups
                sleeperMatchupIdsForThisWeek = {
                    sleeperMatchup.matchup_id for sleeperMatchup in sleeperMatchupsForThisWeek
                }
                if self.__isCompletedWeek(weekNumber, sleeperLeague):
                    # sort matchups by roster IDs
                    rosterIdToSleeperMatchupMap: dict[int, SleeperMatchup] = dict()
                    for sleeperMatchup in sleeperMatchupsForThisWeek:
                        rosterIdToSleeperMatchupMap[sleeperMatchup.roster_id] = sleeperMatchup
                    for sleeperPlayoffMatchup in playoffRoundAndSleeperPlayoffMatchups[roundNumber]:
                        # check if this matchup is in this week (needed for leagues with multiple weeks in a single round)
                        if (
                            sleeperPlayoffMatchup.matchup_id in sleeperMatchupIdsForThisWeek
                            or sleeperLeague.settings.playoff_round_type_enum
                            == SleeperPlayoffRoundType.ONE_WEEK_PER_ROUND
                        ):
                            # team A
                            teamARosterId = sleeperPlayoffMatchup.team_1_roster_id
                            teamA = self.__sleeperRosterIdToTeamMap[teamARosterId]
                            teamAPoints = rosterIdToSleeperMatchupMap[teamARosterId].points
                            teamAHasTiebreaker = (
                                sleeperPlayoffMatchup.winning_roster_id
                                == sleeperPlayoffMatchup.team_1_roster_id
                            )
                            # team B
                            teamBRosterId = sleeperPlayoffMatchup.team_2_roster_id
                            teamB = self.__sleeperRosterIdToTeamMap[teamBRosterId]
                            teamBPoints = rosterIdToSleeperMatchupMap[teamBRosterId].points
                            teamBHasTiebreaker = (
                                sleeperPlayoffMatchup.winning_roster_id
                                == sleeperPlayoffMatchup.team_2_roster_id
                            )

                            multiWeekMatchupId = None
                            # determine if this is a championship matchup or not
                            matchupType = MatchupType.PLAYOFF
                            if sleeperPlayoffMatchup.p == 1:
                                matchupType = MatchupType.CHAMPIONSHIP
                                if (
                                    sleeperLeague.settings.playoff_round_type_enum
                                    == SleeperPlayoffRoundType.TWO_WEEK_CHAMPIONSHIP_ROUND
                                ):
                                    multiWeekMatchupId = f"{teamA.id}{teamB.id}"
                            if (
                                sleeperLeague.settings.playoff_round_type_enum
                                == SleeperPlayoffRoundType.TWO_WEEKS_PER_ROUND
                            ):
                                multiWeekMatchupId = f"{teamA.id}{teamB.id}"
                            matchups.append(
                                Matchup(
                                    teamAId=teamA.id,
                                    teamBId=teamB.id,
                                    teamAScore=teamAPoints,
                                    teamBScore=teamBPoints,
                                    teamAHasTiebreaker=teamAHasTiebreaker,
                                    teamBHasTiebreaker=teamBHasTiebreaker,
                                    matchupType=matchupType,
                                    multiWeekMatchupId=multiWeekMatchupId,
                                )
                            )
                    weeks.append(Week(weekNumber=weekNumber, matchups=matchups))
        return weeks

    def __yearHasDivisions(self, sleeperLeague: SleeperLeague) -> bool:
        return sleeperLeague.settings.divisions is not None

    def __isCompletedWeek(self, weekNumber: int, sleeperLeague: SleeperLeague) -> bool:
        # see if this is the current year/week of the NFL
        sportState = self.__getSleeperSportState()
        return not (
            sportState.season == sleeperLeague.season
            and sportState.leg <= weekNumber
            and sleeperLeague.status != SleeperSeasonStatus.COMPLETE
        )

    def __buildTeams(self, sleeperLeague: SleeperLeague) -> list[Team]:
        teams = list()
        sleeperUsers = self.__getSleeperUsers(sleeperLeague.league_id)
        sleeperRosters = LeagueAPIClient.get_rosters(league_id=sleeperLeague.league_id)
        for sleeperUser in sleeperUsers:
            # connect a sleeperUser to a sleeperRoster
            rosterId = None
            divisionId = None
            for sleeperRoster in sleeperRosters:
                if sleeperRoster.owner_id == sleeperUser.user_id:
                    rosterId = sleeperRoster.roster_id
                    if self.__yearHasDivisions(sleeperLeague):
                        divisionId = self.__sleeperDivisionIdToDivisionMap[
                            sleeperRoster.settings.division
                        ].id
            if rosterId is None:
                raise DoesNotExistException(
                    f"No Roster ID match found for Sleeper User with ID: '{sleeperUser.user_id}'."
                )
            owner = self.__sleeperUserIdToOwnerMap[sleeperUser.user_id]
            teamName = sleeperUser.display_name
            # if we can find a team name for this user, use it instead of their display name
            if (
                isinstance(sleeperUser.metadata, dict)
                and "team_name" in sleeperUser.metadata.keys()
            ):
                teamName = sleeperUser.metadata["team_name"]
            team = Team(ownerId=owner.id, name=teamName, divisionId=divisionId)
            teams.append(team)
            self.__sleeperRosterIdToTeamMap[rosterId] = team
        return teams

    def __loadOwners(self, sleeperLeagues: list[SleeperLeague]) -> None:
        for sleeperLeague in sleeperLeagues:
            sleeperUsers = self.__getSleeperUsers(sleeperLeague.league_id)
            for sleeperUser in sleeperUsers:
                ownerName = sleeperUser.display_name
                # get general owner name if there is one
                generalOwnerName = self._getGeneralOwnerNameFromGivenOwnerName(ownerName)
                ownerName = generalOwnerName if generalOwnerName is not None else ownerName
                self.__sleeperUserIdToOwnerMap[sleeperUser.user_id] = Owner(name=ownerName)

    @staticmethod
    def __calculate_number_of_playoff_weeks(
        sleeperLeague: SleeperLeague, sleeperPlayoffMatchups: list[SleeperPlayoffMatchup]
    ) -> int:
        # sort sleeperPlayoffMatchups by round into a dict
        playoffRoundAndSleeperPlayoffMatchups: dict[int, list[SleeperPlayoffMatchup]] = dict()
        for sleeperPlayoffMatchup in sleeperPlayoffMatchups:
            if sleeperPlayoffMatchup.round in playoffRoundAndSleeperPlayoffMatchups.keys():
                playoffRoundAndSleeperPlayoffMatchups[sleeperPlayoffMatchup.round].append(
                    sleeperPlayoffMatchup
                )
            else:
                playoffRoundAndSleeperPlayoffMatchups[sleeperPlayoffMatchup.round] = [
                    sleeperPlayoffMatchup
                ]
        numberOfPlayoffRounds = max(
            [playoffMatchup.round for playoffMatchup in sleeperPlayoffMatchups]
        )  # don't know a better way to determine this
        match sleeperLeague.settings.playoff_round_type_enum:
            case SleeperPlayoffRoundType.ONE_WEEK_PER_ROUND:
                return numberOfPlayoffRounds
            case SleeperPlayoffRoundType.TWO_WEEK_CHAMPIONSHIP_ROUND:
                return numberOfPlayoffRounds + 1
            case SleeperPlayoffRoundType.TWO_WEEKS_PER_ROUND:
                return numberOfPlayoffRounds * 2
            case _:
                raise LeagueLoaderException(
                    f"PlayoffRoundType '{sleeperLeague.settings.playoff_round_type_enum}' is not supported."
                )

    @staticmethod
    def __create_playoff_week_round_list(
        sleeperLeague: SleeperLeague, playoffWeeks: list, numberOfPlayoffRounds: int
    ) -> list[tuple[int, int]]:
        match sleeperLeague.settings.playoff_round_type_enum:
            case SleeperPlayoffRoundType.ONE_WEEK_PER_ROUND:
                return list(zip(playoffWeeks, range(1, numberOfPlayoffRounds + 1)))
            case SleeperPlayoffRoundType.TWO_WEEK_CHAMPIONSHIP_ROUND:
                return list(
                    itertools.zip_longest(
                        playoffWeeks,
                        range(1, numberOfPlayoffRounds + 1),
                        fillvalue=numberOfPlayoffRounds,
                    )
                )
            case SleeperPlayoffRoundType.TWO_WEEKS_PER_ROUND:
                playoffRounds = [
                    playoffRound
                    for playoffRound in range(1, numberOfPlayoffRounds + 1)
                    for _ in range(1, numberOfPlayoffRounds + 1)
                ]
                return list(zip(playoffWeeks, playoffRounds))



import unittest

from leeger.calculator.year_calculator import ScoringStandardDeviationYearCalculator
from leeger.enum.MatchupType import MatchupType
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.util.Deci import Deci
from test.helper.prototypes import getNDefaultOwnersAndTeams


class TestScoringStandardDeviationYearCalculator(unittest.TestCase):
    def test_getScoringStandardDeviation_happyPath(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.2,
            teamBScore=2.5,
            matchupType=MatchupType.PLAYOFF,
        )
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=5,
            matchupType=MatchupType.CHAMPIONSHIP,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(year)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0.08164965809277260327324280249"), response[teams[0].id])
        self.assertEqual(Deci("1.202774570177914372863325978"), response[teams[1].id])

    def test_getScoringStandardDeviation_multiWeekMatchups(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.1,
            teamBScore=2.4,
            multiWeekMatchupId="1",
        )
        matchup2 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.2,
            teamBScore=2.5,
            multiWeekMatchupId="1",
        )
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=5,
            matchupType=MatchupType.CHAMPIONSHIP,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(year)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0.5"), response[teams[0].id])
        self.assertEqual(Deci("0.05"), response[teams[1].id])

    def test_getScoringStandardDeviation_noneIfNoGamesPlayed(self):
        owners, teams = getNDefaultOwnersAndTeams(3)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(teamAId=teams[1].id, teamBId=teams[2].id, teamAScore=1.2, teamBScore=2.5)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])

        year = Year(yearNumber=2000, teams=teams, weeks=[week1, week2])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, weekNumberEnd=1
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(3, len(response.keys()))
        self.assertEqual(Deci("0"), response[teams[0].id])
        self.assertEqual(Deci("0"), response[teams[1].id])
        self.assertIsNone(response[teams[2].id])

    def test_getScoringStandardDeviation_onlyPostSeasonIsTrue(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.2,
            teamBScore=2.5,
            matchupType=MatchupType.PLAYOFF,
        )
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=3,
            matchupType=MatchupType.CHAMPIONSHIP,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, onlyPostSeason=True
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0.05"), response[teams[0].id])
        self.assertEqual(Deci("0.25"), response[teams[1].id])

    def test_getScoringStandardDeviation_onlyRegularSeasonIsTrue(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=5, teamBScore=3)
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=2.6,
            matchupType=MatchupType.PLAYOFF,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, onlyRegularSeason=True
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("1.95"), response[teams[0].id])
        self.assertEqual(Deci("0.3"), response[teams[1].id])

    def test_getScoringStandardDeviation_onlyChampionshipIsTrue(self):
        owners, teams = getNDefaultOwnersAndTeams(6)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.3, teamBScore=2.3)
        matchup2 = Matchup(teamAId=teams[2].id, teamBId=teams[3].id, teamAScore=3.3, teamBScore=4.3)
        matchup3 = Matchup(teamAId=teams[4].id, teamBId=teams[5].id, teamAScore=4.3, teamBScore=5.3)
        week1 = Week(weekNumber=1, matchups=[matchup1, matchup2, matchup3])

        matchup4 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=2.3,
            matchupType=MatchupType.PLAYOFF,
        )
        matchup5 = Matchup(
            teamAId=teams[2].id,
            teamBId=teams[3].id,
            teamAScore=3.3,
            teamBScore=4.3,
            matchupType=MatchupType.PLAYOFF,
        )
        matchup6 = Matchup(
            teamAId=teams[4].id,
            teamBId=teams[5].id,
            teamAScore=4.3,
            teamBScore=5.3,
            matchupType=MatchupType.CHAMPIONSHIP,
        )
        week2 = Week(weekNumber=2, matchups=[matchup4, matchup5, matchup6])

        year = Year(yearNumber=2002, teams=teams, weeks=[week1, week2])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, onlyChampionship=True
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(6, len(response.keys()))
        self.assertIsNone(response[teams[0].id])
        self.assertIsNone(response[teams[1].id])
        self.assertIsNone(response[teams[2].id])
        self.assertIsNone(response[teams[3].id])
        self.assertEqual(Deci("0"), response[teams[4].id])
        self.assertEqual(Deci("0"), response[teams[5].id])

    def test_getScoringStandardDeviation_weekNumberStartGiven(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.2, teamBScore=2.5)
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=5.5,
            teamBScore=2.6,
            matchupType=MatchupType.PLAYOFF,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, weekNumberStart=2
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("2.15"), response[teams[0].id])
        self.assertEqual(Deci("0.05"), response[teams[1].id])

    def test_getScoringStandardDeviation_weekNumberEndGiven(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.9, teamBScore=5)
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=2.6,
            matchupType=MatchupType.PLAYOFF,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, weekNumberEnd=2
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0.4"), response[teams[0].id])
        self.assertEqual(Deci("1.3"), response[teams[1].id])

    def test_getScoringStandardDeviation_weekNumberStartGivenAndWeekNumberEndGiven(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup1 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.1, teamBScore=2.4)
        matchup2 = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1.2, teamBScore=2.5)
        matchup3 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.3,
            teamBScore=12.6,
            matchupType=MatchupType.PLAYOFF,
        )
        matchup4 = Matchup(
            teamAId=teams[0].id,
            teamBId=teams[1].id,
            teamAScore=1.4,
            teamBScore=2.7,
            matchupType=MatchupType.PLAYOFF,
        )

        week1 = Week(weekNumber=1, matchups=[matchup1])
        week2 = Week(weekNumber=2, matchups=[matchup2])
        week3 = Week(weekNumber=3, matchups=[matchup3])
        week4 = Week(weekNumber=4, matchups=[matchup4])

        year = Year(yearNumber=2000, teams=[teams[0], teams[1]], weeks=[week1, week2, week3, week4])

        response = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(
            year, weekNumberStart=2, weekNumberEnd=3
        )

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0.05"), response[teams[0].id])
        self.assertEqual(Deci("5.05"), response[teams[1].id])



from typing import Any, Optional
from leeger.util.ConfigReader import ConfigReader

from leeger.util.CustomLogger import CustomLogger
from leeger.util.Deci import Deci


class GeneralUtil:
    @staticmethod
    def filter(*, value: Any, list_: list) -> list:
        """
        Filters out the given value from the given list.
        """
        return [v for v in list_ if v is not value]

    @staticmethod
    def warnForUnusedKwargs(
        kwargs: dict[str, Any], *, excludeKeys: Optional[list[str]] = None
    ) -> None:
        """
        Logs a warning for each present kwarg in the given dict.
        Will ignore kwargs with a key in excludeKeys.
        """
        LOGGER = CustomLogger.getLogger()

        excludeKeys = excludeKeys if excludeKeys is not None else list()
        # get the list of common kwargs that sometimes linger in kwargs to be passed down and used later.
        defaultExcludeKeys = ConfigReader.get("KWARGS", "WARNING_EXCLUDE_KEYS", asType=list)
        excludeKeys += defaultExcludeKeys

        unused_kwargs = [kwarg for kwarg in kwargs.keys() if kwarg not in excludeKeys]
        for kwarg in unused_kwargs:
            LOGGER.warning(f"Keyword argument '{kwarg}' unused.")

    @staticmethod
    def safeSum(*numbers) -> Optional[Deci | int | float]:
        """
        Safely adds numbers where None is ignored.
        Examples:
            * 1 + None = 1
            * None + 1 = 1
            * None + None = None
            * 1 + 1 = 2
        """
        safeSum = None
        for number in numbers:
            if None in (safeSum, number):
                safeSum = number if safeSum is None else safeSum
            else:
                safeSum += number

        return safeSum

    def findDifferentFields(
        dict1: dict,
        dict2: dict,
        *,
        parentKey: str = "",
        ignoreKeyNames: Optional[list[str]] = None,
        isRootDict: bool = True,
    ) -> list[str]:
        """
        Finds the differences in the dicts and returns them.
        Uses dot notation, i.e. 'foo.baz.bar'
        """
        differentFields = []

        if ignoreKeyNames is None:
            ignoreKeyNames = []

        for key in dict1:
            fullKey = f"{parentKey}.{key}" if parentKey else key

            value1 = dict1[key]
            value2 = dict2[key]

            if isinstance(value1, dict) and isinstance(value2, dict):
                isRootDict = False
                differentFields.extend(
                    GeneralUtil.findDifferentFields(
                        value1,
                        value2,
                        parentKey=fullKey,
                        ignoreKeyNames=ignoreKeyNames,
                        isRootDict=isRootDict,
                    )
                )
            elif isinstance(value1, (list, tuple)) and isinstance(value2, (list, tuple)):
                if len(value1) != len(value2):
                    differentFields.append((fullKey, (value1, value2)))
                else:
                    list_differences = []
                    for i in range(len(value1)):
                        if isinstance(value1[i], dict) and isinstance(value2[i], dict):
                            nested_diff = GeneralUtil.findDifferentFields(
                                value1[i],
                                value2[i],
                                parentKey=f"{fullKey}[{i}]",
                                ignoreKeyNames=ignoreKeyNames,
                                isRootDict=False,
                            )
                            differentFields.extend(nested_diff)
                        elif value1[i] != value2[i]:
                            list_differences.append(i)
                    if len(list_differences) == len(value1):
                        differentFields.append((fullKey, (value1, value2)))
                    elif list_differences:
                        for index in list_differences:
                            differentFields.append(
                                (f"{fullKey}[{index}]", (value1[index], value2[index]))
                            )
            elif value1 != value2 and key not in ignoreKeyNames:
                differentFields.append((fullKey, (value1, value2)))

        return differentFields



import unittest

from leeger.enum.MatchupType import MatchupType
from leeger.model.filter.AllTimeFilters import AllTimeFilters


class TestAllTimeFilters(unittest.TestCase):
    def test_includeMatchupTypes_default(self):
        allTimeFilters = AllTimeFilters(
            yearNumberStart=2000,
            yearNumberEnd=2001,
            weekNumberStart=1,
            weekNumberEnd=2,
            onlyChampionship=False,
            onlyPostSeason=False,
            onlyRegularSeason=False,
        )

        self.assertEqual(
            [MatchupType.REGULAR_SEASON, MatchupType.PLAYOFF, MatchupType.CHAMPIONSHIP],
            allTimeFilters.includeMatchupTypes,
        )

    def test_includeMatchupTypes_onlyChampionshipIsTrue(self):
        allTimeFilters = AllTimeFilters(
            yearNumberStart=2000,
            yearNumberEnd=2001,
            weekNumberStart=1,
            weekNumberEnd=2,
            onlyChampionship=True,
            onlyPostSeason=False,
            onlyRegularSeason=False,
        )

        self.assertEqual([MatchupType.CHAMPIONSHIP], allTimeFilters.includeMatchupTypes)

    def test_includeMatchupTypes_onlyPostSeasonIsTrue(self):
        allTimeFilters = AllTimeFilters(
            yearNumberStart=2000,
            yearNumberEnd=2001,
            weekNumberStart=1,
            weekNumberEnd=2,
            onlyChampionship=False,
            onlyPostSeason=True,
            onlyRegularSeason=False,
        )

        self.assertEqual(
            [MatchupType.PLAYOFF, MatchupType.CHAMPIONSHIP], allTimeFilters.includeMatchupTypes
        )

    def test_includeMatchupTypes_onlyRegularSeasonIsTrue(self):
        allTimeFilters = AllTimeFilters(
            yearNumberStart=2000,
            yearNumberEnd=2001,
            weekNumberStart=1,
            weekNumberEnd=2,
            onlyChampionship=False,
            onlyPostSeason=False,
            onlyRegularSeason=True,
        )

        self.assertEqual([MatchupType.REGULAR_SEASON], allTimeFilters.includeMatchupTypes)
